/* Copyright (C) Sportradar AG. See LICENSE for full license governing this code */
package com.sportradar.unifiedodds.example.sender;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.concurrent.TimeoutException;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;

import org.xml.sax.InputSource;

import com.rabbitmq.client.Channel;
import com.rabbitmq.client.ConnectionFactory;
import com.sportradar.uf.datamodel.UFBetCancel;
import com.sportradar.uf.datamodel.UFBetSettlement;
import com.sportradar.uf.datamodel.UFBetStop;
import com.sportradar.uf.datamodel.UFFixtureChange;
import com.sportradar.uf.datamodel.UFOddsChange;
import com.sportradar.uf.datamodel.UFRollbackBetCancel;
import com.sportradar.uf.datamodel.UFRollbackBetSettlement;

/**
 * The following example is really not an example but a tool to send a large number of previously
 * recorded odds feed messages to your own local RabbitMQ Server. This can be used for you to local
 * test and replay and stress your own client.
 * 
 * How to run this: java CannedSequenceSender <file-with-a-long-sequence-of-xml-messages>
 * 
 * it takes four optional arguments as system properties -Dhost=
 * <your-rabbit-mq-host-defaults-to-localhost> -Dspeedup=
 * <how-much-faster-to-replay-than-the-original-defaults-to-60x-faster> -Dmax_delay=
 * <longest_delay_between_messages-in-seconds-defaults-to-1> -Dexchange=
 * <name-of-mq-exchange-to-publish-on-defaults-to-srexchange>
 */
public class CannedSequenceSender {

    private static int speedup;
    private static int maxDelay;
    private static String host;
    private static String exchange;

    public static void main(String[] args) throws IOException, JAXBException, InterruptedException, TimeoutException {
        CannedSequenceSender f = new CannedSequenceSender();
        host = System.getProperty("host", "localhost");
        speedup = Integer.parseInt(System.getProperty("speedup", "60"));
        maxDelay = Integer.parseInt(System.getProperty("max_delay", "1"));
        exchange = System.getProperty("exchange", "srexchange");
        f.sendMessages(new File(args[0]));
    }

    private void sendMessages(File file) throws IOException, JAXBException, InterruptedException, TimeoutException {
        BufferedReader br = new BufferedReader(new FileReader(file));
        String msg = "";
        Unmarshaller m = null;
        try {
            JAXBContext context = JAXBContext.newInstance("com.sportradar.uf.datamodel");
            m = context.createUnmarshaller();
        } catch (JAXBException e) {
            br.close();
            throw new IllegalStateException(MessageFormat
                        .format("Failed to create unmarshaller for data-model {0}. Message:{1}", "cp", e.getMessage()));
        }
        ConnectionFactory factory = new ConnectionFactory();
        factory.setHost(host);
        com.rabbitmq.client.Connection con = factory.newConnection();
        Channel channel = con.createChannel();
        channel.exchangeDeclare(exchange, "topic");
        long prevTime = 0L;
        System.out.println("Sending messages now.");
        while (true) {
            String line = br.readLine();
            if (line == null)
                break;
            if (line.startsWith("<?xml")) {

                if (msg.length() != 0) {
                    long time = extractTimestampFromMessage(msg, m, prevTime);
                    long diff = time - prevTime;
                    if (diff < 0)
                        diff = 0;
                    prevTime = time;
                    diff = diff / speedup;
                    if (diff > maxDelay * 1000)
                        diff = maxDelay * 1000;

                    Thread.sleep(diff);

                    channel.basicPublish(exchange, "1.2.3.4", null, msg.getBytes());
                }
                msg = line;
            } else {
                msg += line.trim();
            }

        }
        System.out.println("All messages sent.");
        br.close();
    }

    private long extractTimestampFromMessage(String msg, Unmarshaller m, long prevTime) throws JAXBException {
        long time = prevTime;
        Object o = m.unmarshal(new InputSource(new ByteArrayInputStream(msg.getBytes())));

        if (o instanceof UFOddsChange) {
            time = ((UFOddsChange) o).getTimestamp();
        } else if (o instanceof UFBetCancel) {
            time = ((UFBetCancel) o).getTimestamp();
        } else if (o instanceof UFBetSettlement) {
            time = ((UFBetSettlement) o).getTimestamp();
        } else if (o instanceof UFBetStop) {
            time = ((UFBetStop) o).getTimestamp();
        } else if (o instanceof UFFixtureChange) {
            time = ((UFFixtureChange) o).getTimestamp();
        } else if (o instanceof UFRollbackBetSettlement) {
            time = ((UFRollbackBetSettlement) o).getTimestamp();
        } else if (o instanceof UFRollbackBetCancel) {
            time = ((UFRollbackBetCancel) o).getTimestamp();
        }
        return time;
    }

}
