/* Copyright (C) Sportradar AG. See LICENSE for full license governing this code */
package com.sportradar.unifiedodds.restfuloddscache;

import java.util.List;

import com.sportradar.utils.Cache;

/**
 * This class is the implementation of the MBean for the RestfulOddsCache that provides valuable
 * statistics about the OddsCache
 */
public class OddsCacheStats implements OddsCacheStatsMBean {
    private OddsListener listener;
    private Cache<String, List<byte[]>> logs;
    private long lastCalculatedTime;
    private int logsSize;
    private int prematchMatches;
    private int liveMatches;
    private int finishedMatches;
    private int activeMatches;
    private int unsettledMatches;
    private int inChangesList;
    private int endedMatches;

    public OddsCacheStats(OddsListener listener, Cache<String, List<byte[]>> lm) {
        logs = lm;
        this.listener = listener;
    }

    /**
     * This is a helper function to avoid too frequent recomputation of expensive to compute
     * statistics It caches the results and avoids recomputation for 10 seconds. These means the
     * more expensive counters are not updated more frequently than every 10 secs.
     */
    private synchronized void computeStats() {
        long now = System.currentTimeMillis();
        long diff = now - lastCalculatedTime;
        if (diff < 1000 * 10) {
            return;
        }
        lastCalculatedTime = now;
        logsSize = logs.size(s -> s.stream().mapToInt(a -> a.length).sum());
        listener.fillInStats(this);
    }

    /**
     * This is the call back called by the OddsFeedListener.fillInStats that feeds back the
     * statistics values of interest.
     * 
     * @param prematch
     * @param live
     * @param ended
     * @param finished
     * @param active
     * @param unsettled
     * @param inChanges
     */
    public void setStats(int prematch, int live, int ended, int finished, int active, int unsettled, int inChanges) {
        prematchMatches = prematch;
        liveMatches = live;
        endedMatches = ended;
        finishedMatches = finished;
        activeMatches = active;
        unsettledMatches = unsettled;
        inChangesList = inChanges;
    }

    @Override
    public int getNumberOfUnprocessedMessageInChangesQueue() {
        return inChangesList;
    }

    @Override
    public int getNumberOfMatches() {
        return prematchMatches + liveMatches + endedMatches + finishedMatches;
    }

    @Override
    public int getNumberOfPrematchMatches() {
        computeStats();
        return prematchMatches;
    }

    @Override
    public int getNumberOfLiveMatches() {
        computeStats();
        return liveMatches;
    }

    @Override
    public int getNumberOfFinishedMatches() {
        computeStats();
        return finishedMatches;
    }

    @Override
    public int getNumberOfUnsettledMatches() {
        computeStats();
        return unsettledMatches;
    }

    @Override
    public int getNumberOfActiveMatches() {
        computeStats();
        return activeMatches;
    }

    @Override
    public int getLogSize() {
        computeStats();
        return logsSize;
    }

}
