/* Copyright (C) Sportradar AG. See LICENSE for full license governing this code */
package com.sportradar.unifiedodds.restfuloddscache;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sportradar.unifiedodds.sdk.OddsFeedListener;
import com.sportradar.unifiedodds.sdk.OddsFeedSession;
import com.sportradar.unifiedodds.sdk.impl.sports.EventImpl;
import com.sportradar.unifiedodds.sdk.oddsentities.BetCancel;
import com.sportradar.unifiedodds.sdk.oddsentities.BetSettlement;
import com.sportradar.unifiedodds.sdk.oddsentities.BetStop;
import com.sportradar.unifiedodds.sdk.oddsentities.FixtureChange;
import com.sportradar.unifiedodds.sdk.oddsentities.Market;
import com.sportradar.unifiedodds.sdk.oddsentities.MarketBetSettlement;
import com.sportradar.unifiedodds.sdk.oddsentities.MarketOddsChange;
import com.sportradar.unifiedodds.sdk.oddsentities.MarketStatus;
import com.sportradar.unifiedodds.sdk.oddsentities.OddsChange;
import com.sportradar.unifiedodds.sdk.oddsentities.OutcomeOdds;
import com.sportradar.unifiedodds.sdk.oddsentities.OutcomeResult;
import com.sportradar.unifiedodds.sdk.oddsentities.RollbackBetCancel;
import com.sportradar.unifiedodds.sdk.oddsentities.RollbackBetSettlement;
import com.sportradar.unifiedodds.sdk.sportentities.SportEvent;
import com.sportradar.unifiedodds.sdk.sportentities.SportEventStatus;
import com.sportradar.utils.Cache;

public class OddsListener implements OddsFeedListener {
	private static final int MAX_STORED_EVENT_IDS = 50;

	private Map<String, MatchOdds> top = new HashMap<>();
	private List<MatchOdds> changesList = new LinkedList<>();
	private Set<String> lastEventIds = new HashSet<>();
	private Cache<String, List<byte[]>> logMap;

	private Logger logger = LoggerFactory.getLogger("roc");

	OddsListener(Cache<String, List<byte[]>> lm) {
		logMap = lm;
	}

	@Override
	public void onOddsChange(OddsFeedSession sender, OddsChange<SportEvent> oddsChanges) {
		logRawMessage(oddsChanges.getEvent().getId(), oddsChanges.getRawMessage());
		MatchOdds ofm = getLatestOddsForMatch(oddsChanges.getEvent());
		long now = System.currentTimeMillis();
		synchronized (ofm) {
			ofm.mark();
			for (MarketOddsChange ocfm : oddsChanges.getOddsChanges()) {
				MarketOdds mo = ofm.get(ocfm.getMarket());
				if (mo == null) {
					mo = new MarketOdds(ocfm.getMarket());
					mo.setStatus(MarketStatus.Suspended, oddsChanges.getTimestamp(), oddsChanges.getGeneratingProduct(), ocfm.isFavourite());
					ofm.addMarketOdds(ocfm.getMarket(), mo);
				}
				MarketStatus oldStatus = mo.getStatus();
				MarketStatus newStatus = ocfm.getStatus();
				if (newStatus == MarketStatus.HandedOver) {
					if (mo.getLastProduct() != oddsChanges.getGeneratingProduct()) {
						// We have already received odds from another odds producer ignore this
						// update
						continue;
					}
					else {
						// OK we will soon receive an odds update for this market from another
						// producer - let's betstop it for now
						newStatus = MarketStatus.Suspended;
					}
				}
				mo.setStatus(newStatus, now, oddsChanges.getGeneratingProduct(), ocfm.isFavourite());
				if (newStatus == MarketStatus.Deactivated) {
					continue;
				}
				if (oldStatus == MarketStatus.Settled) {
					if (newStatus != MarketStatus.Settled) {
						logger.warn("Market was settled but is now getting info "
								+ oddsChanges.getEvent().getId()
								+ " "
								+ mo.getMarket().getId());
					}
					continue;
				}
				for (OutcomeOdds ocfo : ocfm.getOutcomeOdds()) {
					if (!ocfo.isActive()) {
						mo.remove(ocfo.getOutcome());
						mo.removeProbs(ocfo.getOutcome());
					}
					mo.set(ocfo.getOutcome(), ocfo.getOdds());
					mo.setProbs(ocfo.getOutcome(), ocfo.getProbability());
				}
			}
		}
	}

	@Override
	public void onBetStop(OddsFeedSession sender, BetStop<SportEvent> betStop) {
		MatchOdds ofm = getLatestOddsForMatch(betStop.getEvent());
		logRawMessage(betStop.getEvent().getId(), betStop.getRawMessage());
		long now = System.currentTimeMillis();
		synchronized (ofm) {
			for (MarketOdds mo : ofm.getMarkets()) {
				mo.setStatus(MarketStatus.Deactivated, now, betStop.getGeneratingProduct());
			}
		}
	}

	@Override
	public void onBetSettlement(OddsFeedSession sender, BetSettlement<SportEvent> clearBets) {
		MatchOdds ofm = getLatestOddsForMatch(clearBets.getEvent());
		logRawMessage(clearBets.getEvent().getId(), clearBets.getRawMessage());
		long now = System.currentTimeMillis();
		for (MarketBetSettlement bsfm : clearBets.getBetSettlements()) {
			Market m = bsfm.getMarket();
			MarketOdds mo = ofm.get(m);
			if (mo != null) {
				mo.setStatus(MarketStatus.Settled, now, clearBets.getGeneratingProduct());
				for (OutcomeResult or : bsfm.getOutcomeResults()) {
					if (or.isWinning())
						mo.setResult(or.getOutcome(), 1.0);
					else
						mo.setResult(or.getOutcome(), 0.0);
				}
			}
		}
	}

	@Override
	public void onRollbackBetSettlement(OddsFeedSession sender, RollbackBetSettlement<SportEvent> rollbackBetSettlement) {
		logRawMessage(rollbackBetSettlement.getEvent().getId(), rollbackBetSettlement.getRawMessage());
		long now = System.currentTimeMillis();
		MatchOdds ofm = getLatestOddsForMatch(rollbackBetSettlement.getEvent());
		for (Market mm : rollbackBetSettlement.getMarkets()) {
			MarketOdds mo = ofm.get(mm);
			if (mo != null) {
				mo.setStatus(MarketStatus.Deactivated, now, rollbackBetSettlement.getGeneratingProduct());
			}
		}
	}

	@Override
	public void onBetCancel(OddsFeedSession sender, BetCancel<SportEvent> betCancel) {
		logRawMessage(betCancel.getEvent().getId(), betCancel.getRawMessage());
	}

	@Override
	public void onRollbackBetCancel(OddsFeedSession sender, RollbackBetCancel<SportEvent> rbBetCancel) {
		logRawMessage(rbBetCancel.getEvent().getId(), rbBetCancel.getRawMessage());
	}

	@Override
	public void onFixtureChange(OddsFeedSession sender, FixtureChange<SportEvent> fixtureChange) {
		logRawMessage(fixtureChange.getEvent().getId(), fixtureChange.getRawMessage());
	}

	private void logRawMessage(String id, byte[] tmpMsg) {
		synchronized (logMap) {
			if (lastEventIds.size() == MAX_STORED_EVENT_IDS)
				lastEventIds.remove(0);
			lastEventIds.add(id);
			List<byte[]> ls = logMap.get(id);
			if (ls == null) {
				ls = new LinkedList<>();
				logMap.put(id, ls);
			}
			ls.add(tmpMsg);
		}
	}

	MatchOdds getNextChange() {
		synchronized (top) {
			if (changesList.size() == 0)
				return null;
			MatchOdds mo = changesList.get(0);
			changesList.remove(0);
			return mo;
		}
	}

	private MatchOdds getLatestOddsForMatch(SportEvent event) {
		MatchOdds ofm;
		synchronized (top) {
			ofm = top.get(event.getId());
			if (ofm == null) {
				ofm = new MatchOdds(event);
				top.put(event.getId(), ofm);
			}
			if (!changesList.contains(ofm)) {
				changesList.add(ofm);
			}
		}
		return ofm;
	}

	public MatchOdds getOddsForMatch(String id) {
		synchronized (top) {
			return top.get(id);
		}
	}

	public String getLog(String id) {
		synchronized (logMap) {
			List<byte[]> log = logMap.get(id);
			if (log == null)
				return "";
			StringBuilder sb = new StringBuilder();
			for (byte[] msg : log) {
				String lines[] = new String(msg).split("\\r?\\n");
				for (String s : lines) {
					if (s.indexOf("xml") != -1)
						continue;
					sb.append(s);
					sb.append('\n');
				}
			}
			return sb.toString();
		}
	}

	public void fillInStats(OddsCacheStats stats) {
		synchronized (top) {
			int inChanges = changesList.size();
			int prematch = 0;
			int live = 0;
			int finished = 0;
			int unsettled = 0;
			int active = 0;
			int ended = 0;
			long now = System.currentTimeMillis();
			for (MatchOdds mo : top.values()) {
				SportEventStatus stat = ((EventImpl)mo.getSportEvent()).getStatusAsync();
				if (stat == null) {
					stat = SportEventStatus.NotStarted;
				}
				long diff = now - mo.getLastChangeTimestamp();
				if (diff < 1000 * 60 * 10) {
					// changes in the last 10 minutes means active
					active++;
				}
				switch (stat) {
					case NotStarted:
						prematch++;
						break;
					case Live:
					case Suspended:
						live++;
						break;
					case Ended:
						ended++;
						break;
					case Cancelled:
					case Finished:
						finished++;
						// need to check for unsettled here
						int aCount = 0;
						int bCount = 0;
						for (MarketOdds mmo : mo.getMarkets()) {
							aCount++;
							if (mmo.getStatus() != MarketStatus.Settled) {
								bCount++;
							}
						}
						if (aCount == -1)
							System.exit(0);
						if (bCount > 0)
							unsettled++;
						break;
				}
			}
			stats.setStats(prematch, live, ended, finished, active, unsettled, inChanges);
		}
	}

	public List<MatchOdds> getUnsettledEvents() {
		List<MatchOdds> ls = new LinkedList<>();
		synchronized (top) {
			for (MatchOdds mo : top.values()) {
				SportEventStatus stat = ((EventImpl)mo.getSportEvent()).getStatusAsync();
				if (stat != null && stat == SportEventStatus.Finished) {
					for (MarketOdds mmo : mo.getMarkets()) {
						if (mmo.getStatus() != MarketStatus.Settled) {
							ls.add(mo);
							break;
						}
					}
				}
			}
		}
		return ls;
	}

	public Set<String> getRecentlyChangedEventIds() {
		synchronized (logMap) {
			return new HashSet<>(lastEventIds);
		}
	}

	public List<Integer> getLogQueueAges() {
		return logMap.getMaxAgeEntries();
	}

	@Override
	public void onUnparseableMessage(OddsFeedSession sender, byte[] rawMessage, SportEvent event) {
		// do nothing
	}
}
